/*  $Id: vsprintf.cc,v 1.5 1998/08/22 03:35:37 shaggy Exp $
 *      var-arg printf to a string buffer
 *      relative path: ./lib/posix
 *  
 *      Shag/OS: an Object-Oriented MicroKernel Operating System
 *       Libraries
 *        Posix-compliant code
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the Shag/OS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

//  $Log: vsprintf.cc,v $
//  Revision 1.5  1998/08/22 03:35:37  shaggy
//  header update
//
//  Revision 1.4  1997/06/21 12:24:20  shaggy
//  header update
//
//  Revision 1.3  1997/01/21 05:03:11  shaggy
//  fixed for inttypes.h
//  removed sprintf
//
//  Revision 1.2  1996/12/20  18:10:51  shaggy
//  internalized long types
//
//  Revision 1.1  1996/12/08  18:37:34  shaggy
//  Initial revision
//
//      creation date: 1992


#include <types.h>
#include <posix/stdarg.h>
#include <posix/stddef.h>
#include <posix/stdio.h>

/*  the following format characters are supported:
 *	%d	decimal
 *	%u	unsigned decimal
 *	%o	unsigned octal
 *	%x	unsigned hex
 *	%b	unsigned binary
 *	%X	capital unsigned hex
 *	%s	string
 *	%c	character
 *	%n	the number of characters written is stored
 *		in the integer pointed to by the field
 * an optional field width can be specified after the %,
 * and a leading zero implies padding with zeros
 * a negative field width will left justify the field
 */

typedef long long		llong;
typedef unsigned long long	ullong;
typedef long double		ldouble;


typedef const char *base_t;
#define BASE(b)		(b)[0]
#define DIGIT(b,x)	(b)[(x)+1]

static const base_t base_binary= "\00201";
static const base_t base_octal = "\01001234567";
static const base_t base_decimal = "\0120123456789";
static const base_t base_hex = "\0200123456789abcdef";
static const base_t base_HEX = "\0200123456789ABCDEF";


/* sprint_ullong()  takes a string buffer, a number, and a base-string,
 *		 and prints the number to the string, and returns
 *	         the number of digits
 * XXX - can be majorly optimized by providing separate
 * 	 functions for each of the bases, or for smaller
 *	 number sizes
 */

static int sprint_ullong(char *s, ullong num, base_t base)
{
register ullong q, r;
register int len;
	/* a minor attempt at optimization without resorting to separate functions... */
	switch(BASE(base)) {
	case 2: q = num/2; r = num%2; break;
	case 8: q = num/8; r = num%8; break;
	case 10: q = num/10; r = num%10; break;
	case 16: q = num/16; r = num%16; break;
	/* shouldn't be any others... but just in case... */
	default: q = 0; r = 0;
	}
	if (q > 0)
		len = sprint_ullong(s, q, base);
	else
		len = 0;
	s[len] = DIGIT(base, r);
	return len+1;
} 


static int sprint_llong(char *s, llong num, base_t base)
{
	if(num < 0) {
		s[0] = '-';
		return sprint_ullong(&s[1], -num, base)+1;
	} else
		return sprint_ullong(s, num, base);
}



/* vsprintf()   takes a string buffer, a format string
 *  		and a pointer to a list of arguments
 *		which are written to the buffer
 *		it returns the number of characters actually written
 */


int vsprintf(char *s, const char *fmt, va_list args)
{
register char *sp;
register int len;
register int n;
ullong arg;
double dfarg;
ldouble xfarg;
bool left;
bool is_half;
bool is_long;
bool is_llong;
bool is_alternate;
char tmp[22];	/* currently, the longest possible number, plus a null */
char *tp;
char padchar;
int i;
base_t base;
	len = 0;
	while(*fmt) if (*fmt == '%') {
		if(fmt[1]=='%') {
			fmt++;
			*s++ = *fmt++;
			continue;
		}	
		if(fmt[1]=='/') {	
			fmt += 2;
			switch(*fmt++) {
			case 'f':
				*s++ = '\f'; break;
			case 'a':
				*s++ = '\007'; break;
			case 't':
				*s++ = '\t'; break;
			case 'b':
				*s++ = '\b'; break;
			case 'e':
				*s++ = '\033'; break;
			case 'r':
				*s++ = '\r'; break;
			case 'n':
				*s++ = '\n'; break;
			}
			continue;
		}
		++fmt;
 		n = 0;
		if((is_alternate = (*fmt == '#')))
			fmt++;
		left = false;
		if (*fmt == '-') {
			left = true;
			fmt++;
		}
		if (*fmt == '0') 
			padchar = '0';
		else
			padchar = ' '; 
		while(*fmt >= '0' && *fmt <= '9') 
			n = n * 10 + (*fmt++ - '0');
		sp = tp = tmp;
		tmp[0] = '\0';
		is_half = is_long = is_llong = false;
		while(true) {
		    switch(*fmt++) {
		    case 'h':
			is_half = true;
			continue;
		    case 'L':
			is_long = true;
		    case 'l':
			is_llong = is_long;
			is_long = true;
			continue;
	    	    case 's':
			sp = va_arg(args, char*);
			if(!sp)
				sp = "(null)";
			break;
		    case 'c':
			tmp[0] = va_arg(args, char);
			tmp[1] = '\0';
			break;
		    /* Yes, look everyone, I'm using 'goto' statements!
		     * Make fun of them, laugh, criticize my code, etc,
		     * all you want, but please don't bother telling me
		     * about it until you can offer me a solution that
		     * is equally efficient for this code. (I know of
		     * several ways to avoid this, but they involve either
		     * a lot of unnecessary code duplication, or in having
		     * a second switch statement.  Both solutions are
		     * unnecessarily inefficient, either in size or speed)
		     */
		    case 'b': base = base_binary; goto print_unum;
		    case 'o': base = base_octal; goto print_unum;
			if(is_alternate) *tp++ = '0'; 
			goto print_unum;
		    case 'x': base = base_hex; 
			if(is_alternate) { *tp++ = '0'; *tp++ = 'x'; }
			goto print_unum;
		    case 'X': base = base_HEX;
			if(is_alternate) { *tp++ = '0'; *tp++ = 'X'; }
			goto print_unum;
		    case 'u': base = base_decimal;
		    print_unum:
			if(is_llong) arg = va_arg(args, ullong);
			else if(is_long) arg = va_arg(args, ulong);
			else if(is_half) arg = va_arg(args, ushort);
			else arg = va_arg(args, uint);
			tp[sprint_ullong(tp, arg, base)] = '\0';
			break;
		    case 'd': case 'i':
			if(is_llong) arg = va_arg(args, llong);
			else if(is_long) arg = va_arg(args, long);
			else if(is_half) arg = va_arg(args, short);
			else arg = va_arg(args, int);
			tmp[sprint_llong(tmp, arg, base_decimal)] = '\0';
			break;
		    case 'p':
			tmp[0] = '0';
			tmp[1] = 'x';
			tmp[sprint_ullong(&tmp[2],
			   (ullong)(ulong)va_arg(args, void*), base_hex)+2] = '\0';
			break;
		    case 'n':
			if(is_llong) *va_arg(args, llong*) = len;
			else if(is_long) *va_arg(args, long*) = len;
			else if(is_half) *va_arg(args, short*) = len;
			else *va_arg(args, int*) = len;
			break;
		    /* XXX - floating point is not yet supported... */
		    /* More 'goto's... see note above... */
		    case 'e': /* set some flag */; goto print_float;
		    case 'E': /* set some flag */; goto print_float;
		    case 'f': /* set some flag */; goto print_float;
		    case 'g': /* set some flag */; goto print_float;
		    print_float:
			if(is_llong) xfarg = va_arg(args, ldouble);
			else dfarg = va_arg(args, double);
			sp = "(n/a)";
			break;
		    default:
			sp = "(?)";
		    }
		    break;
		}
		if (!left) {
			i = n - strlen(sp);
			while(i-- > 0) {
				*s++ = padchar;
				len++;
			}
		}
		i = n;
	   	while(*sp && (!n || (i > 0))) {
			*s++ = *sp++;
			len++;
			i--;
		}
		while(left && (i-- > 0)) {
			*s++ = ' ';
			len++;
		}
	} else {
		*s++ = *fmt++;
		len++;
	}
	*s = '\0';
	return(len);
}


